/*
 * Copyright (C) 2002-2003 Stefan Stiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.kiwisoft.scrap;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Observer;
import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.wm.ToolWindow;
import com.intellij.openapi.wm.ToolWindowManager;

import com.kiwisoft.utils.DocumentAdapter;
import com.kiwisoft.utils.NotifyObject;
import com.kiwisoft.utils.StringUtils;
import com.kiwisoft.utils.Utils;
import com.kiwisoft.utils.gui.IconManager;
import com.kiwisoft.utils.gui.TristateBoxIcon;
import com.kiwisoft.utils.gui.lookup.DialogLookupField;

/**
 * @author Stefan Stiller
 * @version $Revision: 1.3 $, $Date: 2004/10/20 17:45:06 $
 */
public class MainPanel extends JPanel
{
	private Project project;

	private JTabbedPane tabs;
	private DialogLookupField tfImports;

	private CopyAction copyAction;

	public MainPanel(final Project project)
	{
		// Copies the input map from TextArea to get the same action keys as in the editor window
		UIManager.put("EditorPane.focusInputMap", UIManager.get("TextArea.focusInputMap"));
		UIManager.put("EditorPane.font", UIManager.get("TextArea.font"));


		UIManager.put("TristateBoxUI", "com.kiwisoft.utils.gui.TristateBoxUI");
		UIManager.put("TristateBox.icon", new TristateBoxIcon());
		UIManager.put("TristateBox.focus", UIManager.getColor("CheckBox.focus"));
		UIManager.put("TristateBox.disabledText", UIManager.getColor("CheckBox.disabledText"));
		UIManager.put("TristateBox.focusInputMap", new UIDefaults.LazyInputMap(new Object[]{
			"SPACE", "pressed",
			"released SPACE", "released"
		}));

		this.project=project;
		setLayout(new BorderLayout());

		tfImports=new DialogLookupField(new ImportsLookup(project));
		tfImports.setText(ScrapConfig.getInstance(project).getImports());
		tfImports.getDocument().addDocumentListener(new DocumentAdapter()
		{
			public void changedUpdate(DocumentEvent e)
			{
				ScrapConfig.getInstance(project).setImports(tfImports.getText());
			}
		});

		JPanel pnlImports=new JPanel(new GridBagLayout());
		pnlImports.add(new JLabel("Main Imports:"), new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0,
				GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0, 0, 0,0), 0, 0));
		pnlImports.add(tfImports, new GridBagConstraints(1, 0, 1, 1, 1.0, 0.0,
				GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(0, 5, 0,0), 0, 0));

		JToolBar toolbar=new JToolBar();
		toolbar.setFloatable(false);
		toolbar.add(Utils.createButton(new CreateAction()));
		copyAction=new CopyAction();
		toolbar.add(Utils.createButton(copyAction));
		toolbar.addSeparator();
		toolbar.add(pnlImports);
		toolbar.addSeparator();
		toolbar.add(Utils.createButton(new HelpAction()));
		toolbar.add(Utils.createButton(new AboutAction(this)));

		tabs=new JTabbedPane();
		tabs.setTabPlacement(JTabbedPane.BOTTOM);
		tabs.addChangeListener(new ScrapSelectionListener());

		add(toolbar, BorderLayout.NORTH);
		add(tabs, BorderLayout.CENTER);

		ScrapManager scrapManager=ScrapManager.getInstance(project);
		scrapManager.addObserver(new ScrapManagerObserver());

		List scrapList=scrapManager.getScraps();
		if (scrapList.isEmpty())
			scrapManager.createScrap();
		else
		{
			Iterator it=scrapList.iterator();
			while (it.hasNext()) addScrap((Scrap)it.next());
		}
		updateActions();

		tabs.addMouseListener(new TabsMouseListener());
	}

	private Scrap getSelectedScrap()
	{
		if (tabs!=null)
		{
			ScrapPanel scrapPanel=(ScrapPanel)tabs.getSelectedComponent();
			if (scrapPanel!=null) return scrapPanel.getScrap();
		}
		return null;
	}

	private ScrapPanel getScrapPanel(Scrap scrap)
	{
		Component[] panels=tabs.getComponents();
		for (int i=0; i<panels.length; i++)
		{
			if (panels[i] instanceof ScrapPanel)
			{
				ScrapPanel scrapPanel=(ScrapPanel)panels[i];
				if (scrapPanel.getScrap()==scrap) return scrapPanel;
			}
		}
		return null;
	}

	private void addScrap(Scrap newScrap)
	{
		ScrapPanel scrapPanel=new ScrapPanel(project, this, newScrap);
		tabs.add(getTabName(newScrap), scrapPanel);
		tabs.setSelectedComponent(scrapPanel);
	}

	private class ScrapSelectionListener implements ChangeListener
	{
		public void stateChanged(ChangeEvent e)
		{
			updateActions();
		}
	}

	private static String getTabName(Scrap scrap)
	{
		return scrap.getName();
	}

	private class ScrapManagerObserver implements Observer
	{
		public void update(Observable o, Object arg)
		{
			NotifyObject note=(NotifyObject)arg;
			Object type=note.getArgument(0);
			if ("scrap added".equals(type))
			{
				Scrap newScrap=(Scrap)note.getArgument(1);
				addScrap(newScrap);
				updateActions();
			}
			else if ("scrap removed".equals(type))
			{
				Scrap scrap=(Scrap)note.getArgument(1);
				ScrapPanel scrapPanel=getScrapPanel(scrap);
				if (scrapPanel!=null) tabs.remove(scrapPanel);
				updateActions();
				if (tabs.getTabCount()==0 && project!=null)
				{
					ToolWindow toolWindow=ToolWindowManager.getInstance(project).getToolWindow(ScrapPlugin.TOOL_WINDOW_ID);
					if (toolWindow!=null) toolWindow.hide(null);
				}
			}
			else if ("scrap changed".equals(type) || "query name changed".equals(type))
			{
				Scrap scrap=(Scrap)note.getArgument(1);
				ScrapPanel scrapPanel=getScrapPanel(scrap);
				if (scrapPanel!=null)
				{
					int index=tabs.indexOfComponent(scrapPanel);
					tabs.setTitleAt(index, getTabName(scrap));
				}
				updateActions();
			}
		}
	}

	private void updateActions()
	{
		Scrap selectedScrap=getSelectedScrap();
		boolean scrapSelected=selectedScrap!=null;
		if (copyAction!=null) copyAction.setEnabled(scrapSelected);
	}

	public String[] getImports()
	{
		String importText=tfImports.getText().trim();
		if (!StringUtils.isEmpty(importText))
		{
			return importText.split(";|,");
		}
		return null;
	}

	public ScrapPanel getSelectedPanel()
	{
		Component selectedComponent=tabs.getSelectedComponent();
		if (selectedComponent instanceof ScrapPanel) return (ScrapPanel)selectedComponent;
		return null;
	}

	private class CreateAction extends AbstractAction
	{
		public CreateAction()
		{
			super(null, IconManager.getIcon(MainPanel.class.getClassLoader(), "icons/add.png"));
			putValue(Action.SHORT_DESCRIPTION, "Open a new empty scrap.");
		}

		public void actionPerformed(ActionEvent e)
		{
			ScrapManager.getInstance(project).createScrap();
		}
	}

	private class CopyAction extends AbstractAction
	{
		public CopyAction()
		{
			super(null, IconManager.getIcon(MainPanel.class.getClassLoader(), "icons/copy.png"));
			putValue(Action.SHORT_DESCRIPTION, "Create a new scrap and copy the text from the selected.");
		}

		public void actionPerformed(ActionEvent e)
		{
			Scrap scrap=getSelectedScrap();
			if (scrap!=null) ScrapManager.getInstance(project).createScrap(scrap);
		}
	}

	private class TabsMouseListener extends MouseAdapter
	{
		public void mousePressed(MouseEvent e)
		{
			if ((e.getModifiers()&MouseEvent.BUTTON3_MASK)!=0)
			{
				int index=tabs.getUI().tabForCoordinate(tabs, e.getX(), e.getY());
				if (index>=0)
				{
					Component component=tabs.getComponentAt(index);
					if (component instanceof ScrapPanel)
					{
						ScrapPanel panel=(ScrapPanel)component;
						JPopupMenu popup=new JPopupMenu();
						popup.add(new JMenuItem(new RenameAction(panel.getScrap())));
						popup.show(tabs, e.getX(), e.getY()-20);
					}
				}
			}
		}
	}

	private class RenameAction extends AbstractAction
	{
		private Scrap scrap;

		public RenameAction(Scrap scrap)
		{
			super("Rename scrap");
			this.scrap=scrap;
			setEnabled(scrap!=null);
		}

		public void actionPerformed(ActionEvent e)
		{
			RenameDialogStub dialogStub=new RenameDialogStub("Rename Scrap", scrap.getName());
			Utils.wrapDialog(project, dialogStub, true, true);
			String newName=dialogStub.getReturnValue();
			if (!StringUtils.isEmpty(newName)) scrap.setName(newName);
		}
	}

}
