/*
 * Copyright (C) 2002-2003 Stefan Stiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.kiwisoft.scrap;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.*;
import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;
import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.text.*;

import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.ModuleManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.ModuleRootManager;
import com.intellij.openapi.roots.OrderEntry;
import com.intellij.openapi.roots.OrderRootType;
import com.intellij.openapi.roots.ProjectRootManager;
import com.intellij.openapi.vfs.VirtualFile;

import com.kiwisoft.utils.DocumentAdapter;
import com.kiwisoft.utils.StringUtils;
import com.kiwisoft.utils.Utils;
import com.kiwisoft.utils.format.FormatManager;
import com.kiwisoft.utils.gui.ConfirmationDialogStub;
import com.kiwisoft.utils.gui.IconManager;
import com.kiwisoft.utils.gui.inspect.InspectorDialogStub;
import com.kiwisoft.utils.gui.lookup.DialogLookupField;
import com.kiwisoft.utils.gui.text.SourceTextPane;
import com.kiwisoft.utils.gui.text.SyntaxDefinitionFactory;
import net.sf.pizzacompiler.compiler.*;

/**
 * @author Stefan Stiller
 * @version $Revision: 1.5 $, $Date: 2004/10/20 17:45:34 $
 */
public class ScrapPanel extends JPanel
{
	public static final int ERROR=1;
	public static final int OUTPUT=2;
	public static final int RESULT=3;
	public static final int INFO=4;
	public static final int DEFAULT=0;

	private Project project;
	private MainPanel mainPanel;
	private Scrap scrap;
	private ScrapResult scrapResult;
	private DefaultComboBoxModel modules=new DefaultComboBoxModel();

	private JTextComponent tfScrap;
	private JTextPane tfResult;
	private DialogLookupField tfImports;
	private StatusBar pnlStatus;

	private ScrapDocumentListener scrapDocumentListener;

	private RunAction runAction;
	private EvaluateAction evaluateAction;
	private InspectAction inspectAction;

	private ExecutionThread executionThread;
	private JComboBox cbxModules;

	public ScrapPanel(Project project, MainPanel mainPanel, Scrap scrap)
	{
		this.project=project;
		this.mainPanel=mainPanel;
		this.scrap=scrap;
		setLayout(new BorderLayout());

		runAction=new RunAction();
		evaluateAction=new EvaluateAction();
		inspectAction=new InspectAction();

		tfScrap=new SourceTextPane(SyntaxDefinitionFactory.getInstance().getSyntaxDefinition("java"));
		tfScrap.setText(scrap.getText());
		tfImports=new DialogLookupField(new ImportsLookup(project));
		tfImports.setText(scrap.getImports());

		scrapDocumentListener=new ScrapDocumentListener();
		tfScrap.getDocument().addDocumentListener(scrapDocumentListener);
		tfImports.getDocument().addDocumentListener(scrapDocumentListener);

		if (project!=null)
		{
			ModuleManager moduleManager=ModuleManager.getInstance(project);
			Module[] modules=moduleManager.getSortedModules();
			for (int i=0; i<modules.length; i++) this.modules.addElement(modules[i]);
		}
		cbxModules=new JComboBox(modules);
		cbxModules.setRenderer(new ModuleRenderer());

		JPanel pnlButtons=new JPanel(new GridBagLayout());
		pnlButtons.add(Utils.createButton(runAction),
				new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Utils.createButton(evaluateAction),
				new GridBagConstraints(0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Utils.createButton(inspectAction),
				new GridBagConstraints(0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Utils.createButton(new CancelAction()),
				new GridBagConstraints(0, 3, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlButtons.add(Box.createGlue(),
				new GridBagConstraints(0, 6, 1, 1, 0.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));

		tfResult=new JTextPane();
		Style style=tfResult.addStyle("error", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.RED);
		style=tfResult.addStyle("info", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.GREEN);
		style=tfResult.addStyle("default", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.BLACK);
		style=tfResult.addStyle("output", null);
		style.addAttribute(StyleConstants.ColorConstants.Foreground, Color.BLACK);
		style=tfResult.addStyle("result", null);
		style.addAttribute(StyleConstants.ColorConstants.Bold, Boolean.TRUE);

		pnlStatus=new StatusBar();
		pnlStatus.setPreferredSize(new Dimension(20, 20));

		JPanel pnlTop=new JPanel(new GridBagLayout());
		pnlTop.add(new JLabel("Modules:"),
				new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlTop.add(cbxModules,
				new GridBagConstraints(1, 0, 1, 1, 0.5, 0.0, GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(2, 2, 2, 2), 0, 0));
		pnlTop.add(new JLabel(" Imports:"),
				new GridBagConstraints(2, 0, 1, 1, 0.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));
		pnlTop.add(tfImports,
				new GridBagConstraints(3, 0, 1, 1, 0.5, 0.0, GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(2, 2, 2, 2), 0, 0));

		JSplitPane splitPane=new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		splitPane.setOneTouchExpandable(true);
		splitPane.setTopComponent(new JScrollPane(tfScrap));
		splitPane.setBottomComponent(new JScrollPane(tfResult));
		splitPane.setDividerLocation(100);

		JPanel pnlMain=new JPanel(new BorderLayout());
		pnlMain.add(pnlTop, BorderLayout.NORTH);
		pnlMain.add(splitPane, BorderLayout.CENTER);
		pnlMain.add(pnlStatus, BorderLayout.SOUTH);

		add(pnlButtons, BorderLayout.WEST);
		add(pnlMain, BorderLayout.CENTER);
		if (project!=null)
		{
			try
			{
				Class aClass=Class.forName("com.kiwisoft.scrap.ModuleAdapter");
				Constructor constructor=aClass.getConstructor(new Class[]{Project.class, JComboBox.class});
				constructor.newInstance(new Object[]{project, cbxModules});
			}
			catch (Throwable e)
			{
				System.out.println("Module listener not initialized.");
			}
		}
	}

	public void addNotify()
	{
		super.addNotify();
		validateActions();
	}

	public Scrap getScrap()
	{
		return scrap;
	}

	private void validateActions()
	{
		if (runAction!=null) runAction.validate();
		if (evaluateAction!=null) evaluateAction.validate();
		if (inspectAction!=null) inspectAction.validate();
	}

	private class ScrapDocumentListener extends DocumentAdapter
	{
		public void changedUpdate(DocumentEvent e)
		{
			scrap.setImports(tfImports.getText());
			scrap.setText(tfScrap.getText());
			validateActions();
		}
	}

	private String getScrapText()
	{
		pnlStatus.showMessage("");

		String fullText=tfScrap.getText();
		String partialText=tfScrap.getSelectedText();

		String scrapText=partialText;

		if (StringUtils.isEmpty(scrapText))
			scrapText=fullText;
		else
		{
			ScrapConfig config=ScrapConfig.getInstance(project);
			Boolean confirm=config.getConfirmation("partialExecute");
			if (Boolean.FALSE.equals(confirm))
			{
				scrapText=fullText;
			}
			else if (confirm==null)
			{
				ConfirmationDialogStub dialog=new ConfirmationDialogStub("Question", "Do you want to execute only the selected part of the scrap?",
						new String[]{"Yes", "No", "Cancel"}, "Yes", true);
				Utils.wrapDialog(project, dialog, true, true);
				if ("Yes".equals(dialog.getReturnValue()))
				{
					if (dialog.isDontShowAgain()) config.setConfirmation("partialExecute", Boolean.TRUE);
				}
				else if ("No".equals(dialog.getReturnValue()))
				{
					if (dialog.isDontShowAgain()) config.setConfirmation("partialExecute", Boolean.FALSE);
					scrapText=fullText;
				}
				else
					return null;
			}
		}

		return scrapText.trim();
	}

	protected void evaluateScrap()
	{
		if (executionThread==null || !executionThread.isAlive())
		{
			String scrapText=getScrapText();
			if (scrapText!=null)
			{
				String[] lines=scrapText.split("\r|\n");
				StringBuffer scrapCode=new StringBuffer();
				for (int i=0; i<lines.length; i++)
				{
					String line=lines[i];
					if (i==lines.length-1)
						scrapCode.append("return $result("+line+");");
					else
						scrapCode.append(line);
					scrapCode.append("\n");
				}
				executeScrap(scrapCode.toString());
			}
		}
	}

	protected void runScrap()
	{
		if (executionThread==null || !executionThread.isAlive())
		{
			String scrapText=getScrapText();
			if (scrapText!=null)
			{
				executeScrap(scrapText+"return null;");
			}
		}
	}

	private void executeScrap(String scrapCode)
	{
		synchronized (ExecutionThread.class)
		{
			try
			{
				net.sf.pizzacompiler.compiler.Main.init();
				ClassLoader classLoader;
				ClassReader classReader;
				if (project==null)
				{
					classLoader=Thread.currentThread().getContextClassLoader();
					classReader=new ClassReader();
				}
				else
				{
					Module module=(Module)cbxModules.getSelectedItem();
					Set projectClassPath;
					if (module!=null)
						projectClassPath=getModuleClassPath(module);
					else
						projectClassPath=getProjectClassPath();

					StringBuffer classPath=new StringBuffer();
					URL[] urls=new URL[projectClassPath.size()];
					int i=0;
					for (Iterator it=projectClassPath.iterator(); it.hasNext(); i++)
					{
						String pathElement=(String)it.next();
						classPath.append(pathElement).append(File.pathSeparator);
						urls[i]=new File(pathElement).toURL();
					}
					classPath.append(Utils.getClassBasePath(FormatManager.class).getAbsolutePath()).append(File.pathSeparator);
					classPath.append(Utils.getClassBasePath(ScrapPlugin.class).getAbsolutePath()).append(File.pathSeparator);
					classPath.append(Utils.getClassBasePath(ClassReader.class).getAbsolutePath()).append(File.pathSeparator);
					System.out.println("classPath = "+classPath);

					classReader=new ClassReader(null, "\""+classPath+"\"");
					classLoader=new URLClassLoader(urls, ScrapBase.class.getClassLoader());
				}
				executionThread=new ExecutionThread(scrapCode, mainPanel.getImports(), getImports(), classLoader, classReader);
				clearLastResults();
				runAction.setEnabled(false);
				evaluateAction.setEnabled(false);
				pnlStatus.startProgress();
				executionThread.start();
			}
			catch (MalformedURLException e)
			{
				handleThrowable(e);
			}
		}
	}

	public String[] getImports()
	{
		String importText=tfImports.getText().trim();
		if (!StringUtils.isEmpty(importText))
		{
			return importText.split(";|,");
		}
		return null;
	}

	private class RunAction extends AbstractAction
	{
		public RunAction()
		{
			super("Run");
			putValue(Action.SMALL_ICON, IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/run.png"));
			putValue(Action.SHORT_DESCRIPTION, "Run scrap.");
			putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.ALT_MASK));
		}

		public void validate()
		{
			setEnabled((executionThread==null || !executionThread.isAlive()) && !StringUtils.isEmpty(tfScrap.getText()));
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				runScrap();
			}
			catch (Exception e1)
			{
				handleThrowable(e1);
			}
		}
	}

	private class EvaluateAction extends AbstractAction
	{
		public EvaluateAction()
		{
			super("Evaluate");
			putValue(Action.SMALL_ICON, IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/evaluate.png"));
			putValue(Action.SHORT_DESCRIPTION, "Evaluate scrap.");
			putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, KeyEvent.CTRL_MASK));
		}

		public void validate()
		{
			setEnabled((executionThread==null || !executionThread.isAlive()) && !StringUtils.isEmpty(tfScrap.getText()));
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				evaluateScrap();
			}
			catch (Exception e1)
			{
				handleThrowable(e1);
			}
		}
	}

	private class InspectAction extends AbstractAction
	{
		public InspectAction()
		{
			super("Inspect");
			putValue(Action.SMALL_ICON, IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/browser.png"));
			putValue(Action.SHORT_DESCRIPTION, "Inspect scrap result.");
		}

		public void validate()
		{
			setEnabled(scrapResult!=null && scrapResult.isObject());
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				if (scrapResult!=null && scrapResult.isObject())
					Utils.wrapDialog(project, new InspectorDialogStub("Inspector", scrapResult.getValue()), false, true);
			}
			catch (Exception e1)
			{
				handleThrowable(e1);
			}
		}
	}

	public void handleThrowable(Throwable e)
	{
		e.printStackTrace();
		showMessage(e.getClass().getName()+": "+e.getMessage(), ERROR, true);
	}

	private void clearLastResults()
	{
		scrapResult=null;
		tfResult.setText("");
	}

	public void showMessage(String message, int type, boolean nextLine)
	{
		StyledDocument document=tfResult.getStyledDocument();
		try
		{
			if (document.getLength()>0)
			{
				if (nextLine)
				{
					String last=document.getText(document.getLength()-1, 1);
					if (!"\n".equals(last))
						document.insertString(document.getLength(), "\n", null);
				}
			}
			switch (type)
			{
				case OUTPUT:
					document.setLogicalStyle(document.getLength(), document.getStyle("output"));
					break;
				case RESULT:
					document.setLogicalStyle(document.getLength(), document.getStyle("result"));
					break;
				case ERROR:
					document.setLogicalStyle(document.getLength(), document.getStyle("error"));
					break;
				case INFO:
					document.setLogicalStyle(document.getLength(), document.getStyle("info"));
					break;
				default:
					document.setLogicalStyle(document.getLength(), document.getStyle("default"));
			}
			document.insertString(document.getLength(), message, null);
			if (nextLine) document.insertString(document.getLength(), "\n", null);
			tfResult.setCaretPosition(document.getLength());
		}
		catch (BadLocationException e)
		{
			e.printStackTrace();
		}
	}

	private class CancelAction extends AbstractAction
	{
		public CancelAction()
		{
			super("Close scrap", IconManager.getIcon(ScrapPanel.class.getClassLoader(), "icons/cancel.png"));
			putValue(Action.SHORT_DESCRIPTION, "Close current scrap.");
		}

		public void actionPerformed(ActionEvent e)
		{
			ScrapManager.getInstance(project).removeScrap(scrap);
		}
	}

	public void insertString(final String text)
	{
		SwingUtilities.invokeLater(new Runnable()
		{
			public void run()
			{
				Document document=tfScrap.getDocument();
				try
				{
					int selectionStart=tfScrap.getSelectionStart();
					int selectionEnd=tfScrap.getSelectionEnd();
					int position=tfScrap.getCaretPosition();
					if (selectionStart<selectionEnd)
					{
						position=selectionStart;
						document.remove(selectionStart, selectionEnd-selectionStart);
					}
					document.insertString(position, text, null);
					tfScrap.requestFocus();
				}
				catch (BadLocationException e)
				{
					e.printStackTrace();
				}
			}
		});
	}

	private Set getProjectClassPath()
	{
		ProjectRootManager rootManager=(ProjectRootManager)project.getComponent(ProjectRootManager.class);
		VirtualFile[] virtualFiles=rootManager.getFullClassPath();
		Set list=new LinkedHashSet();
		int i;
		for (i=0; i<virtualFiles.length; i++)
		{
			String path=virtualFiles[i].getPath();
			list.add(path.endsWith("!/") ? path.substring(0, path.length()-2) : path);
		}
		return list;
	}

	private static Set getModuleClassPath(Module module)
	{
		Set list=new LinkedHashSet();
		ModuleRootManager moduleRoot=(ModuleRootManager)module.getComponent(ModuleRootManager.class);
		OrderEntry[] orderEntries=moduleRoot.getOrderEntries();
		for (int i=0; i<orderEntries.length; i++)
		{
			OrderEntry entry=orderEntries[i];
			VirtualFile[] files=entry.getFiles(OrderRootType.CLASSES_AND_OUTPUT);
			for (int j=0; j<files.length; j++)
			{
				String path=files[j].getPath();
				list.add(path.endsWith("!/") ? path.substring(0, path.length()-2) : path);
			}
		}
		return list;
	}

	private class ExecutionThread extends Thread
	{
		private String text;
		private String[] mainImports;
		private String[] imports;
		private ClassLoader classLoader;
		private ClassReader classReader;

		public ExecutionThread(String text, String[] mainImports, String[] imports, ClassLoader classLoader, ClassReader classReader)
		{
			this.text=text;
			this.mainImports=mainImports;
			this.imports=imports;
			this.classLoader=classLoader;
			this.classReader=classReader;
		}

		public void run()
		{
			long time1=System.currentTimeMillis();
			try
			{
				ByteArrayOutputStream messageStream=new ByteArrayOutputStream();
				ClassLoader scrapLoader;
				try
				{
					String sourceCode=ScrapTemplate.getInstance().buildClass(mainImports, imports, text);

					net.sf.pizzacompiler.compiler.Main.setClassReader(classReader);

					ByteArrayCompilerOutput compilerOutput=new ByteArrayCompilerOutput();
					HashMap classMap=new HashMap();
					classMap.put(ScrapTemplate.CLASS_NAME, sourceCode);
					SourceReader reader=new MapSourceReader(classMap);
					PrintStream sysOut=new PrintStream(messageStream);
					net.sf.pizzacompiler.compiler.Main.compile(new String[]{ScrapTemplate.CLASS_NAME}, reader, compilerOutput, sysOut);
					scrapLoader=new ByteArrayClassLoader(classLoader, compilerOutput.getBytecode());
					if (messageStream.size()>0)
					{
						showMessage("Compiling error:", ERROR, true);
						showMessage(new String(messageStream.toByteArray()), ERROR, true);
						return;
					}
				}
				catch (Throwable throwable)
				{
					handleThrowable(throwable);
					return;
				}
				Writer outputWriter;
				ScrapBase scrap;
				try
				{
					Class aClass=scrapLoader.loadClass(ScrapTemplate.CLASS_NAME);
					scrap=(ScrapBase)aClass.newInstance();
					outputWriter=new MessageWriter();
				}
				catch (Throwable throwable)
				{
					showMessage("Instantiation error:", ERROR, true);
					handleThrowable(throwable);
					return;
				}
				scrapResult=scrap.$scrap(new PrintWriter(outputWriter));
				if (scrapResult!=null) showMessage(scrapResult.toString(), RESULT, true);
				System.out.println();
			}
			catch (Throwable e)
			{
				handleThrowable(e);
			}
			finally
			{
				final long time=System.currentTimeMillis()-time1;
				SwingUtilities.invokeLater(new Runnable()
				{
					public void run()
					{
						pnlStatus.appendMessage("Execution time: "+time+"ms");
						validateActions();
						pnlStatus.stopProgress();
					}
				});
			}
		}

	}

	private class MessageWriter extends Writer
	{
		public void write(char cbuf[], int off, int len)
		{
			showMessage(new String(cbuf, off, len), OUTPUT, false);
		}

		public void flush()
		{
		}

		public void close()
		{
		}
	}
}
